/*****************************************************************************\

Scene.cc
Author: Forrester Cole (fcole@cs.princeton.edu)
Copyright (c) 2009 Forrester Cole

qviewer is distributed under the terms of the GNU General Public License.
See the COPYING file for details.

\*****************************************************************************/

#include "Scene.h"
#include "GLViewer.h"
#include "TriMesh.h"

#include <QFileInfo>
#include <QDir>
#include <QFileDialog>
#include <QMessageBox>
#include <QStringList>

#include "GQDraw.h"
using namespace GQDraw;

#include <assert.h>

const int CURRENT_VERSION = 1;

static dkBool _useSilhouettes("Silhouettes->Activer", false);
static dkFloat _seuil("Silhouettes->Seuil", 0.0005);

static dkBool _useEnvMap("Post-Processing->Env. Map.", false);
static dkBool _useShadows("Post-Processing->Ombres", false);

static dkBool  _useHDR("Tone-mapping->Activer", false);
static dkFloat _gamma("Tone-mapping->Gamma", 1.0, 0.01, 10.0, 0.01);
static dkFloat _exposure("Tone-mapping->Exposition", 1.0, 0.01, 10.0, 0.01);

Scene::Scene()
{
	_trimesh = NULL;
	_shadowTextureSize = 1024;
}

Scene::~Scene()
{
	clear();
	delete _cube;
	delete _sphereBackground;
	delete _sphere;
	delete _quad;
	delete _ground;
	delete _hdrTexReflection;
	delete _tex_earth_day;
	delete _tex_earth_night;
	delete _tex_earth_bump;
	delete _tex_earth_normals;
}

void Scene::init()
{
	// Chargement des textures
	_tex_earth_day   = new GQTexture2D();
	_tex_earth_night = new GQTexture2D();
	_tex_earth_bump = new GQTexture2D();
	_tex_earth_normals = new GQTexture2D();

	if (!_tex_earth_day->load("textures/earth1.png")   || 
		!_tex_earth_night->load("textures/earth2.png") ||
		!_tex_earth_normals->load("textures/earth3.png") ||
		!_tex_earth_bump->load("textures/earth4.png"))
	{
		QMessageBox::critical(NULL, "Open Failed", "Failed to load texture");
	}

	_sphere = new Sphere();
	_sphereBackground = new Sphere(8.f);
	_cube = new Cube();
	_quad = new Quad();
	_ground = new Quad();

	_hdrTexReflection = new GQCubeMap();
	_hdrTexReflection->load("textures/hdr/uffizi_cross2.hdr");
}

void Scene::clear()
{
	delete _trimesh;
	_trimesh = NULL;
	_vertex_buffer_set.clear();
	_viewer_state.clear();
	_dials_and_knobs_state.clear();
}


bool Scene::load( const QString& filename )
{
	if (filename.endsWith(fileExtension()))
	{
		QFile file(filename);
		if (!file.open(QIODevice::ReadOnly))
		{
			qWarning("Could not open %s", qPrintable(filename));
			return false;
		}

		QDomDocument doc("scene");
		QString parse_errors;
		if (!doc.setContent(&file, &parse_errors))
		{
			qWarning("Parse errors: %s", qPrintable(parse_errors));
			return false;
		}

		file.close();

		QDomElement root = doc.documentElement();
		QDir path = QFileInfo(filename).absoluteDir();

		return load(root, path); 
	}
	else
	{
		_trimesh = TriMesh::read(qPrintable(filename));
		_trimesh_filename = filename;
		if (!_trimesh)
		{
			clear();
			return false;
		}
		_viewer_state.clear();
		_dials_and_knobs_state.clear();
		setupMesh();

		return true;
	}
}

bool Scene::load( const QDomElement& root, const QDir& path )
{
	int version = root.attribute("version").toInt();
	if (version != CURRENT_VERSION)
	{
		qWarning("Scene::load: file version out of date (%d, current is %d)", 
			version, CURRENT_VERSION);
		return false;
	}

	QDomElement model = root.firstChildElement("model");
	if (model.isNull())
	{
		qWarning("Scene::load: no model node found.\n");
		return false;
	}

	QString relative_filename = model.attribute("filename");
	QString abs_filename = path.absoluteFilePath(relative_filename);

	_trimesh = TriMesh::read(qPrintable(abs_filename));
	_trimesh_filename = abs_filename;
	if (!_trimesh)
	{
		qWarning("Scene::load: could not load %s\n", 
			qPrintable(abs_filename));
		return false;
	}

	_viewer_state = root.firstChildElement("viewerstate");
	if (_viewer_state.isNull())
	{
		qWarning("Scene::load: no viewerstate node found.\n");
		clear();
		return false;
	}

	_dials_and_knobs_state = root.firstChildElement("dials_and_knobs");
	if (_dials_and_knobs_state.isNull())
	{
		qWarning("Scene::load: no dials_and_knobs node found.\n");
		return false;
	}
	setupMesh();

	return true;
}

bool Scene::save(const QString& filename, const GLViewer* viewer,
				 const DialsAndKnobs* dials_and_knobs)
{
	QDomDocument doc("scene");
	QDomElement root = doc.createElement("scene");
	doc.appendChild(root);

	_viewer_state = viewer->domElement("viewerstate", doc);
	_dials_and_knobs_state = 
		dials_and_knobs->domElement("dials_and_knobs", doc);

	QDir path = QFileInfo(filename).absoluteDir();

	bool ret = save(doc, root, path);
	if (!ret)
		return false;

	QFile file(filename);
	if (!file.open(QIODevice::WriteOnly))
	{
		qWarning("Scene::save - Could not save %s", qPrintable(filename));
		return false;
	}

	file.write(doc.toByteArray());

	file.close();

	return true;
}

bool Scene::save( QDomDocument& doc, QDomElement& root, const QDir& path )
{
	root.setAttribute("version", CURRENT_VERSION);

	QDomElement model = doc.createElement("model");
	model.setAttribute("filename", path.relativeFilePath(_trimesh_filename));
	root.appendChild(model);

	root.appendChild(_viewer_state);
	root.appendChild(_dials_and_knobs_state);

	return true;
}

void Scene::loadEnvMap( const QString &filename )
{
	delete _hdrTexReflection;
	_hdrTexReflection = new GQCubeMap();
	_hdrTexReflection->load(filename);
}

void Scene::boundingSphere(vec& center, float& radius, const ModelType type)
{
	center = vec();
	switch(type)
	{
	case CUBE:
		radius = _cube->radius();
		break;
	case SPHERE:
		radius = _sphere->radius();
		break;
	case MAILLAGE:
		center = _trimesh->bsphere.center;
		radius = _trimesh->bsphere.r;
		if(_useShadows)
			radius *= 3.0f;
		break;
	default:
		break;
	}
	if(_useEnvMap)
	{
		delete _sphereBackground;
		_sphereBackground = new Sphere(8.0f*radius);
		radius = _sphereBackground->radius();
	}
}

void Scene::setupTextures(GQShaderRef& shader)
{
	if(shader.uniformLocation("textureDay")>=0)
		shader.bindNamedTexture("textureDay",_tex_earth_day);
	if(shader.uniformLocation("textureNight")>=0)
		shader.bindNamedTexture("textureNight",_tex_earth_night);
	if(shader.uniformLocation("textureBump")>=0)
		shader.bindNamedTexture("textureBump",_tex_earth_bump);
	if(shader.uniformLocation("textureNormals")>=0)
		shader.bindNamedTexture("textureNormals",_tex_earth_normals);
}

void Scene::drawScene(const QString &programName, const ModelType type)
{
	if (GQShaderManager::status() != GQ_SHADERS_OK)
		return;

	GQShaderRef shader;

	// TP3
	if(_useHDR)
	{
		_fbo.initFullScreen(1,GQ_ATTACH_DEPTH,GQ_COORDS_NORMALIZED);
		_fbo.bind();

		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	}
	if(_useShadows)
	{
		_shadowFBO.init(_shadowTextureSize,_shadowTextureSize,1,GQ_ATTACH_DEPTH_TEXTURE,GQ_COORDS_NORMALIZED,GQ_FORMAT_RGBA_BYTE);
		drawShadowBuffer();
	}
	if(_useEnvMap)
		drawEnvMap();

	// Activation du shader
	shader = GQShaderManager::bindProgram(programName);

	// Envoie des variables uniformes (matrices de transformation)
	shader.setUniformXform("projection_matrix",_projection_matrix);
	shader.setUniformXform("model_view_matrix", _modelView_matrix);
	if(shader.uniformLocation("normal_matrix")>=0)
	{
		xform normal_matrix(_modelView_matrix);
		invert(normal_matrix);
		normal_matrix = transpose(normal_matrix);
		shader.setUniformMatrixUpper3x3("normal_matrix", normal_matrix);
	}

	// TP3
	if(shader.uniformLocation("shadow_matrix")>=0)
	{
		xform shadow_matrix;
		// Biais pour viter le self-shadowing
		xform bias(0.5f, 0.0f, 0.0f, 0.0f,
			0.0f, 0.5f, 0.0f, 0.0f,
			0.0f, 0.0f, 0.5f, 0.0f,
			0.5f, 0.5f, 0.5f, 1.0f);
		shadow_matrix = shadow_matrix * bias;
		
		// Complter les transformations manquantes

		shader.setUniformXform("shadow_matrix", shadow_matrix);
	}

	setupLighting(shader);

	// TP2
	setupTextures(shader);

	// TP3
	if(_useSilhouettes)
	{
		_fbo.initFullScreen(2,GQ_ATTACH_DEPTH_TEXTURE,GQ_COORDS_NORMALIZED,GQ_FORMAT_RGBA_BYTE);
		_fbo.bind();
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	}

	glDepthMask(true);
	glEnable(GL_DEPTH_TEST);
	glDisable(GL_CULL_FACE);

	// TP3
	if(_useEnvMap && shader.uniformLocation("cubemap")>=0)
	{
		shader.bindNamedTexture("cubemap",_hdrTexReflection);
		shader.setUniform4fv("camera_position",_camera_position);
	}
	if(_useShadows)
	{
		glActiveTexture(GL_TEXTURE0);
		glEnable(GL_TEXTURE_2D);
		glBindTexture(GL_TEXTURE_2D,_shadowFBO.depthMapId());
		shader.setUniform1i("shadowTexture",0);
		
		// Sol pour visualiser clairement l'ombre
		drawGround(shader,false);
	}

	switch(type)
	{
	case CUBE:
		_cube->draw(shader);
		break;
	case SPHERE:
		_sphere->draw(shader);
		break;
	case MAILLAGE:
		drawMesh(shader);
		break;
	default:
		break;
	}

	// TP3
	if(_useSilhouettes)
	{
		_fbo.unbind();
		shader.unbind();
		extractSilhouettes();
	}

	if(_useHDR)
	{
		_fbo.unbind();
		shader.unbind();
		toneMap();
	}
	if(_useShadows)
	{
		glActiveTexture(GL_TEXTURE0);
		glDisable(GL_TEXTURE_2D);
	}
}

void Scene::drawGround(GQShaderRef &shader, bool shadowPass)
{
	xform model = xform::trans(0.0,_trimesh->bbox.min[1],-2.0*_trimesh->bsphere.r)
		*xform::rot(M_PI/2.0,1.0,0.0,0.0)
		*xform::scale(3.0*_trimesh->bsphere.r);
	if(shadowPass)
		model = _modelView_shadow_matrix * model;
	else
		model = _modelView_matrix * model;
	shader.setUniformXform("model_view_matrix", model);

	_ground->draw(shader);

	if(shadowPass)
		shader.setUniformXform("model_view_matrix", _modelView_shadow_matrix);
	else
		shader.setUniformXform("model_view_matrix", _modelView_matrix);
}

void Scene::setupMesh()
{
	_trimesh->need_bsphere();
	_trimesh->need_bbox();
	_trimesh->need_normals();
	_trimesh->need_tstrips();
	_trimesh->need_faces();
	_trimesh->need_uv_dirs();
}

void Scene::setupVertexBufferSet()
{
	// Trimesh stores triangle strips as length followed by indices.
	// Grab them and store the offsets and lengths.
	_tristrips.clear();
	const int *t = &_trimesh->tstrips[0];
	const int *end = t + _trimesh->tstrips.size();
	while (t < end) {
		int striplen = *t++;
		_tristrips.push_back(striplen);
		t += striplen;
	}

	_vertex_buffer_set.clear();
	_vertex_buffer_set.add(GQ_VERTEX, _trimesh->vertices);
	_vertex_buffer_set.add(GQ_NORMAL, _trimesh->normals);
	if(!_trimesh->colors.empty())
		_vertex_buffer_set.add(GQ_COLOR,  _trimesh->colors);
	if(!_trimesh->texcoords.empty())
		_vertex_buffer_set.add(GQ_TEXCOORD, _trimesh->texcoords);
	_vertex_buffer_set.add(GQ_INDEX, 1, _trimesh->tstrips);

	_vertex_buffer_set.copyToVBOs();
}	

void Scene::drawMesh(GQShaderRef& shader)
{
	if (_vertex_buffer_set.numBuffers() == 0)
		setupVertexBufferSet();

	assert(_trimesh->tstrips.size() > 0);

	_vertex_buffer_set.bind(shader);

	int offset = 1;
	for (int i = 0; i < _tristrips.size(); i++) {
		drawElements(_vertex_buffer_set, GL_TRIANGLE_STRIP, offset, _tristrips[i]); 
		offset += _tristrips[i] + 1; // +1 to skip the length stored in the tristrip array.
	}

	_vertex_buffer_set.unbind();
}

void Scene::extractSilhouettes()
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glDisable(GL_DEPTH_TEST);
	glDisable(GL_CULL_FACE);

	// Activation du shader d'extraction des silhouettes
	GQShaderRef shader = GQShaderManager::bindProgram("quad");

	glActiveTexture(GL_TEXTURE0);
	glEnable(_fbo.colorTexture(0)->target());
	_fbo.colorTexture(0)->bind();
	shader.setUniform1i("colorMap",0);

	glActiveTexture(GL_TEXTURE1);
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D,_fbo.depthMapId());
	shader.setUniform1i("depthMap",1);

	shader.setUniform1f("width", _fbo.colorTexture(0)->width());
	shader.setUniform1f("height", _fbo.colorTexture(0)->height());
	shader.setUniform1f("threshold", _seuil);

	_quad->draw(shader);

	glActiveTexture( GL_TEXTURE0 );
	glDisable(_fbo.colorTexture(0)->target());

	glActiveTexture( GL_TEXTURE1 );
	glDisable(GL_TEXTURE_2D);

	shader.unbind();
}

void Scene::drawEnvMap()
{
	glCullFace(GL_BACK);
	glEnable(GL_CULL_FACE);

	GQShaderRef shader = GQShaderManager::bindProgram("background");

	// Envoie des variables uniformes (matrices de transformation)
	shader.setUniformXform("projection_matrix",_projection_matrix);
	shader.setUniformXform("model_view_matrix", _modelView_matrix);

	shader.bindNamedTexture("cubemap",_hdrTexReflection);

	if(shader.uniformLocation("normal_matrix")>=0)
	{
		xform normal_matrix(_modelView_matrix);
		invert(normal_matrix);
		normal_matrix = transpose(normal_matrix);
		shader.setUniformMatrixUpper3x3("normal_matrix", normal_matrix);
	}

	_sphereBackground->draw(shader);

	shader.unbind();
}

void Scene::toneMap()
{

}


void Scene::drawShadowBuffer()
{
	// Activation du shader d'ombre
	GQShaderRef shader = GQShaderManager::bindProgram("shadowBuffer");

	qglviewer::Camera lightCam;
	qglviewer::Vec light = qglviewer::Vec(_light_direction[0],_light_direction[1],_light_direction[2]);
	vec center;
	float radius;
	boundingSphere(center,radius,MAILLAGE);
	lightCam.setPosition(light*radius);
	lightCam.lookAt(qglviewer::Vec(center[0],center[1],center[2]));
	lightCam.setSceneCenter(qglviewer::Vec(center[0],center[1],center[2]));
	lightCam.setSceneRadius(radius);
	lightCam.setFOVToFitScene();
	lightCam.setAspectRatio(1);
	lightCam.computeModelViewMatrix();
	lightCam.computeProjectionMatrix();
	lightCam.getModelViewMatrix(_modelView_shadow_matrix);
	lightCam.getProjectionMatrix(_projection_shadow_matrix);

	vec4i coords;
	glGetIntegerv(GL_VIEWPORT, coords);
	glViewport(0, 0, _shadowTextureSize, _shadowTextureSize);

	_shadowFBO.bind();

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glEnable(GL_DEPTH_TEST);

	// Envoie des variables uniformes (matrices de transformation)
	shader.setUniformXform("projection_matrix", _projection_shadow_matrix);
	shader.setUniformXform("model_view_matrix", _modelView_shadow_matrix);

	drawGround(shader,true);
	drawMesh(shader);

	_shadowFBO.unbind();

	shader.unbind();

	// Retour au viewport d'origine
	glViewport(coords[0],coords[1],coords[2],coords[3]);
}


static QStringList lightPresetNames = (QStringList() << "Headlight" <<
									   "North" << "North-Northeast" << "Northeast" << "East-Northeast" <<
									   "East" << "East-Southeast" << "Southeast" << "South-Southeast" <<
									   "South" << "South-Southwest" << "Southwest" << "West-Southwest" <<
									   "West" << "West-Northwest" << "Northwest" << "North-Northwest" );
static dkStringList light_preset("Light->Direction", lightPresetNames);
static dkFloat light_depth("Light->Depth", 1.0f, 0.0f, 2.0f, 0.1f);

void Scene::setupLighting(GQShaderRef& shader)
{
	const QString& preset = light_preset.value();
	vec camera_light;

	if (preset == "Headlight")
		camera_light = vec(0.0f, 0.0f, 1.0f);
	else if (preset == "North")
		camera_light = vec(0.0f, 1.0f, light_depth);
	else if (preset == "North-Northeast")
		camera_light = vec(0.374f, 1.0f, light_depth);
	else if (preset == "Northeast")
		camera_light = vec(1.0f, 1.0f, light_depth);
	else if (preset == "East-Northeast")
		camera_light = vec(1.0f, 0.374f, light_depth);
	else if (preset == "East")
		camera_light = vec(1.0f, 0.0f, light_depth);
	else if (preset == "East-Southeast")
		camera_light = vec(1.0f, -0.374f, light_depth);
	else if (preset == "Southeast")
		camera_light = vec(1.0f, -1.0f, light_depth);
	else if (preset == "South-Southeast")
		camera_light = vec(0.374f, -1.0f, light_depth);
	else if (preset == "South")
		camera_light = vec(0.0f, -1.0f, light_depth);
	else if (preset == "South-Southwest")
		camera_light = vec(-0.374f, -1.0f, light_depth);
	else if (preset == "Southwest")
		camera_light = vec(-1.0f, -1.0f, light_depth);
	else if (preset == "West-Southwest")
		camera_light = vec(-1.0f, -0.374f, light_depth);
	else if (preset == "West")
		camera_light = vec(-1.0f, 0.0f, light_depth);
	else if (preset == "West-Northwest")
		camera_light = vec(-1.0f, 0.374f, light_depth);
	else if (preset == "Northwest")
		camera_light = vec(-1.0f, 1.0f, light_depth);
	else if (preset == "North-Northwest")
		camera_light = vec(-0.374f, 1.0f, light_depth);

	normalize(camera_light);
	xform mv_xf = rot_only( _camera_transform );
	_light_direction = mv_xf * camera_light;

	if(shader.uniformLocation("light_dir_world")>=0)
		shader.setUniform3fv("light_dir_world", _light_direction);
}

void Scene::recordStats(Stats& stats)
{
	stats.beginConstantGroup("Mesh");
	stats.setConstant("Num Vertices", _trimesh->vertices.size());
	stats.setConstant("Num Faces", _trimesh->faces.size());
	stats.endConstantGroup();
}

